# coding: utf-8
"""
Модуль экшенов приложения.
"""
__author__ = 'fattakhov'
import datetime
import logging
import traceback
import tasks
import uuid

from django.db import transaction

from m3.actions import (Action, PreJsonResult, ACD, utils,
                        OperationResult, ControllerCache)
from m3.actions.packs import BaseDictionaryModelActions, DictListWindowAction
from m3.actions.exceptions import ApplicationLogicException
from m3_ext.ui.results import ExtUIScriptResult
from kinder.controllers import obs
from kinder.core.dict.models import UnitKind
from kinder.core.unit.models import Unit
from kinder.core.unit.actions import UnitDictPack
import settings as gs_settings
import forms
from models import (ReportForm, ReportFormRow, ScheduleSettings, TaskScheduler,
                    GisdoUnit)
from helpers import ReportFormHelper, SendReportException

log = logging.getLogger('gisdo')


@obs.subscribe
class AddFieldListener(object):
    """
    Перехватчик создания окна редактирования учреждения,
    добавляем в конец поле Не отправлять фед.отчетность если тип учреждения ДОУ
    """
    listen = ['kinder.core.unit.actions.UnitDictPack/UnitEditNodeWindowAction',
              'kinder.core.unit.actions.UnitDictPack/TreeNewNodeWindowAction']

    def after(self, request, context, result):
        unit_id = getattr(context, 'id', -1)

        extension_win = forms.GisdoUnitEditWindowBuilder.build_extension(result.data)

        try:
            unit = Unit.objects.get(id=unit_id)
            not_on_federal_report = unit.gisdo.not_on_federal_report
        except Unit.DoesNotExist:
            not_on_federal_report = False

        extension_win.not_on_federal_report.checked = not_on_federal_report

        result.data = extension_win

        return result


@obs.subscribe
class SaveFieldListener(object):
    """
    Перехватчик сохранения окна редактирования учреждения,
    сохраняем поле Не отправлять фед.отчетность в модели GisdoUnit
    """
    listen = ['kinder.core.unit.actions.UnitDictPack/TreeSaveNodeAction']

    def after(self, request, context, result):

        unit_id = result.data.get('id', -1)
        unit = Unit.objects.get(id=unit_id)

        not_on_federal_report = False
        if hasattr(context, 'not_on_federal_report') and context.not_on_federal_report == 'on':
            not_on_federal_report = True

        try:
            g_unit = GisdoUnit.objects.get(unit=unit)
        except GisdoUnit.DoesNotExist:
            g_unit = GisdoUnit(unit=unit)
            try:
                g_unit.create_doo_identity()
            except ApplicationLogicException, e:
                Unit.objects.filter(id=unit_id).delete()
                raise ApplicationLogicException(e.exception_message)

        g_unit.not_on_federal_report = not_on_federal_report
        g_unit.save()

        # Возвращаем ответ в исходное положение.
        result.data = None

        return result


class ReportFormActionPack(BaseDictionaryModelActions):
    """
    Экшен пак работы с объектами модели ReportForm
    """
    url = '/report-forms'
    verbose_name = u'Федеральная отчетность'
    title = verbose_name
    model = ReportForm
    list_columns = [
        ('date', u'Дата отчета'),
        ('sent', u'Отправлен'),
        ('user.get_fio', u'Пользователь'),
        ('unit.name', u'Главное учреждение'),
        ('progress', u'Активная сборка'),
    ]
    list_sort_order = ['-date']
    list_form = forms.ReportFormsListWindow
    need_check_permission = True

    #: создание формы
    PERM_ADD = 'add'
    #: удаление формы
    PERM_DELETE = 'delete'
    #: отправка формы на фед.уровень
    PERM_SEND = 'send'
    #: редактирование формы
    PERM_EDIT = 'edit'

    sub_permissions = {
        PERM_ADD:       u'Создание формы',
        PERM_EDIT:      u'Редактирование',
        PERM_DELETE:    u'Удаление формы',
        PERM_SEND:      u'Отправка на Фед.уровень',
    }

    def __init__(self):
        super(ReportFormActionPack, self).__init__()
        self._init_actions()

    def _init_actions(self):
        """
        Инициализация экшенов пака
        """
        #: данные отчета по учреждению
        self.report_form_row_data_action = ReportFormRowDataAction()
        #: запуск таска на подсчет показателей учреждения и его дочерних
        self.report_form_row_collect_action = ReportFormRowCollectAction()
        #: рендер отчета в .xls формат
        self.report_form_row_excel_action = ReportFormRowToExcelAction()
        #: вызов окна сборки отчета
        self.report_form_build_window_action = ReportFormBuildWindowAction()
        #: вызов окна просмотра отчета
        self.report_form_view_window_action = ReportFormViewWindowAction()
        #: отправка отчета
        self.report_form_send_action = ReportFormSentAction()
        #: вызов окна настроек сборки
        self.report_form_settings_window_action = ReportFormSettingsWindowAction()
        #: сохранение настроек сборки
        self.report_form_settings_save_action = ReportFormSettingsSaveAction()

        self.actions.remove(self.list_window_action)
        self.list_window_action = ListWindowAction()
        self.actions.append(self.list_window_action)

        self.actions.extend([
            self.report_form_row_data_action,
            self.report_form_row_collect_action,
            self.report_form_build_window_action,
            self.report_form_view_window_action,
            self.report_form_row_excel_action,
            self.report_form_send_action,
            self.report_form_settings_window_action,
            self.report_form_settings_save_action
        ])

    @property
    def gisdo_pack(self):
        if not hasattr(self, '_gisdo_pack'):
            self._gisdo_pack = ControllerCache.find_pack(GisdoUnitDictPack)
        return self._gisdo_pack

    @property
    def build_window_url(self):
        """URL вызова окна сборки отчета"""
        return self.report_form_build_window_action.get_absolute_url()

    @property
    def save_tree_url(self):
        """URL запуска таска на подсчет показателей"""
        return self.report_form_row_collect_action.get_absolute_url()

    @property
    def settings_window_url(self):
        """URL вызова окна настроек сборки"""
        return self.report_form_settings_window_action.get_absolute_url()

    @property
    def settings_save_url(self):
        """URL сохранения настроек сборки"""
        return self.report_form_settings_save_action.get_absolute_url()

    @property
    def view_window_url(self):
        """URL вызова окна просмотра отчета"""
        return self.report_form_view_window_action.get_absolute_url()

    @property
    def excel_url(self):
        """URL рендера отчета в .xls формат"""
        return self.report_form_row_excel_action.get_absolute_url()

    @property
    def send_url(self):
        """URL отправки отчета"""
        return self.report_form_send_action.get_absolute_url()

    @transaction.commit_on_success
    def delete_row(self, objs):
        """Удаление отчета и всех дочерних ReportFormRow объектов"""
        for obj in objs:
            obj.reportformrow_set.all().delete()
        return super(ReportFormActionPack, self).delete_row(objs)

    def get_rows(self, offset, limit, filter, user_sort=''):

        if user_sort:
            user_sort = user_sort.replace('sent', 'sent_time')
            user_sort = user_sort.replace('progress', 'in_progress')
            user_sort = user_sort.replace('unit.name', 'unit__name')
            user_sort = user_sort.replace('user.get_fio', 'user__fname')

        return super(ReportFormActionPack, self).get_rows(offset, limit, filter, user_sort=user_sort)

    def extend_menu(self, menu):
        return menu.Item(
            name=self.title,
            pack=self.list_window_action
        )


class ListWindowAction(DictListWindowAction):
    need_check_permission = True
    verbose_name = u'Просмотр'

    def run(self, request, context):
        win = self.create_window(request, context, mode=0)
        self.create_columns(win.grid, self.parent.list_columns)
        self.configure_list(win)

        handler_dblclick = win.grid.handler_dblclick
        # проверим право редактирования
        if not self.parent.has_sub_permission(
                request.user, self.parent.PERM_EDIT, request):
            win.make_read_only()
        win.grid.handler_dblclick = handler_dblclick

        return ExtUIScriptResult(self.parent.get_list_window(win))


#-------------------------------------------------
# Генерация SOAP запросов и связанные с этим экшены
#-------------------------------------------------
#-------------------------------------------------
# Экшены ReportFormRow
#-------------------------------------------------
class ReportFormRowDataAction(Action):
    """
    Данные по конкретному учреждению для окна редактирования ReportForm
    """
    url = '/report-form-row-data'
    verbose_name = u'Данные для формы отчета'

    def context_declaration(self):
        return [ACD(name='unit_id', type=int, required=True),
                ACD(name='report_form_id', type=int, required=True)]

    def run(self, request, context):
        """
        Получение данных по id объекта и отправка обратно
        на клиент в формате для отображения
        """
        try:
            report_form_row = ReportFormRow.objects.get(unit__id=context.unit_id,
                                                        report__id=context.report_form_id)
        except ReportFormRow.DoesNotExist:
            return OperationResult.by_message(u'Отсутствуют данные по учреждению')
        else:
            result = {
                'applications': report_form_row.get_applications_view_data(),
                'queues': report_form_row.get_queues_view_data(),
                'enrolled': report_form_row.get_enrolled_view_data(),
                'capacities': report_form_row.get_capacities_view_data(),
            }
            return PreJsonResult(data=result)


class ReportFormRowCollectAction(Action):
    """
    Сохранение данных по учреждению и его дочерним учреждениям
    """
    url = '/save-report-form-row'
    verbose_name = u'Сохранение данных по дереву учреждений'

    def context_declaration(self):
        return ACD(name='unit_id', type=int, required=True),

    def run(self, request, context):
        """
        Создание нового объекта ReportForm и запуск таска на
        подчсет показателей по нему и дочерним учреждениям
        """
        unit = Unit.objects.get(id=context.unit_id)
        if unit.kind.id not in [UnitKind.MO, UnitKind.REGION]:
            return OperationResult.by_message(u'Создание отчета выполняется '
                                              u'с уровня не ниже МО.')

        report_form = ReportForm.objects.create(date=datetime.date.today(),
                                                user=request.user.get_profile(),
                                                unit=unit,
                                                in_progress=True)

        context.report_form = report_form
        tasks.CollectUnitsDataTask.start((context,),
                                         None,
                                         description=u'Подсчет показателей: %s.'
                                                     % report_form.unit.name)

        return OperationResult(message=u"Подсчет показателей начался в фоновом режиме.")


class ReportFormRowToExcelAction(Action):
    """
    Рендер отчета в .xls фомат
    """
    url = '/excel'
    verbose_name = u'Генерация excel'

    def context_declaration(self):
        return [
            ACD(name='report_id', type=int, required=True, verbose_name=u'id отчета'),
            ACD(name='unit_id', type=int, required=True, verbose_name=u'id учреждения')
        ]

    def run(self, request, context):
        """
        Создание отчета в excel средствами
        excel_reporting
        """
        if not gs_settings.REGION_CODE:
            return OperationResult.by_message(u'Не задан код региона')
        result_name = str(uuid.uuid4())[:16]

        full_name = " ".join([request.user.last_name, request.user.first_name])
        email = request.user.email
        phone = ''

        tasks.Report.start((result_name,
                            context.report_id,
                            full_name,
                            email,
                            phone,), None, description=u'Формирование отчета')

        return OperationResult(message=u"Формирование отчета началось в фоновом режиме.")


#-------------------------------------------------
# Экшены ReportForm
#-------------------------------------------------
class ReportFormBuildWindowAction(Action):
    """
    Вызов окна ввода параметров формирования отчета
    """
    url = '/build'
    verbose_name = u'Формирование отчета'

    def run(self, request, context):
        return ExtUIScriptResult(forms.ReportFormBuildWindow(self.parent))


class ReportFormViewWindowAction(Action):
    """
    Окно отображения отчета
    """
    url = '/view'
    verbose_name = u'Просмотр отчета'

    def context_declaration(self):
        return ACD(name='id', type=int, required=True, verbose_name=u'id отчета'),

    def run(self, request, context):
        """
        Окно просмотра отчета не вызывается для отчетов,
        находящихся в процессе сборки
        """
        try:
            report_form = ReportForm.objects.get(id=context.id)
            if not report_form.in_progress:
                win = forms.ReportFormViewWindow(pack=self.parent)
                if not self.parent.has_sub_permission(request.user, self.parent.PERM_SEND, request):
                    win.send_btn.make_read_only()
                result = ExtUIScriptResult(win)
            else:
                result = OperationResult(message=u'Отчет находится в процессе сборки. '
                                                 u'Вы можете увидеть прогресс сборки в меню '
                                                 u'(Администрирование -> Запущенные задачи)')
        except ReportForm.DoesNotExist:
            result = OperationResult.by_message(u'Отчет не найден в базе')

        return result


class ReportFormSentAction(Action):
    """
    Отправка отчета
    """
    url = '/send'
    verbose_name = u'Отправка отчета'

    def context_declaration(self):
        return ACD(name='report_id', required=True, type=int),

    def run(self, request, context):
        """
        Рендер отчета для отправки на веб-сервис в формате xml
        """
        schedule_settings = ScheduleSettings.get_settings()
        report_form = ReportForm.objects.get(id=context.report_id)
        report_helper = ReportFormHelper(report_form, schedule_settings, mode=ReportFormHelper.SEND_MODE)

        report_data = {
            'schema_version': '1.0',
            'report_date': report_helper.report_form.date.date().isoformat()
        }

        error_msg = None

        try:
            client = report_helper.create_client()
        except SendReportException, error:
            return OperationResult.by_message(u'Ошибка при отправке. Сообщение: %s' % error.message)

        for mo in report_helper.mo_list.filter(kind__id=UnitKind.MO):
            report_data['reports'] = []
            for dou in ReportFormHelper.get_list_dou(mo):

                try:
                    sub_report = report_helper.create_sub_report(dou)
                except ApplicationLogicException, e:
                    log.error(traceback.format_exc())
                    error_msg = u'Ошибка при формировании данных по (%s). Сообщение: %s' % (dou.display(),
                                                                                          e.exception_message)
                    break
                if sub_report:
                    report_data['reports'].append(sub_report)

            report_data['send_time'] = datetime.datetime.now().isoformat()
            encoded_xml = report_helper.render_xml(report_data, mo.display())
            try:
                report_helper.send(client, encoded_xml)
            except SendReportException, error:
                log.error(traceback.format_exc())
                error_msg = u'Ошибка при отправке. (%s). Сообщение: %s' % (mo.display(),
                                                                           error.message or '')
                report_helper.report_form.sent_time = None
                break

        report_helper.report_form.in_progress = False

        if not error_msg:
            report_helper.report_form.sent_time = datetime.datetime.now()
            message = u'Отчет успешно отправлен в систему показателей.'
        else:
            message = error_msg

        report_helper.report_form.save()

        return OperationResult(message=message)


class ReportFormSettingsWindowAction(Action):
    """
    Окно отображения настроек сборки отчета
    """
    url = '/settings'
    verbose_name = u'Настройки сбора отчета'

    def run(self, request, context):
        schedule_settings = ScheduleSettings.get_settings()
        window = forms.ReportFormSettingsWindow()
        window.form.from_object(schedule_settings)
        window.form.url = self.parent.settings_save_url
        return ExtUIScriptResult(window)


class ReportFormSettingsSaveAction(Action):
    """
    Сохранение настроек сборки
    """
    url = '/settings-save'
    verbose_name = u'Сохранение настроек сбора отчета'

    def context_declaration(self):
        return ACD(name='id', type=int, verbose_name=u'id'),

    def run(self, request, context):
        """
        В случае соответсвующих настроек создаем
        периодические таски
        """
        get_row = lambda _id: ScheduleSettings.objects.get(id=_id)
        schedule_settings = utils.bind_request_form_to_object(request,
                                                              get_row,
                                                              forms.ReportFormSettingsWindow)
        try:
            schedule_settings.save()
            map(TaskScheduler.terminate, TaskScheduler.objects.all())
            if schedule_settings.is_active:
                ts = TaskScheduler.schedule_every('gisdo.tasks.CollectUnitDataPeriodicTask',
                                                  schedule_settings.time.hour,
                                                  schedule_settings.time.minute)
                ts.start()
        except Exception, error:
            return OperationResult.by_message(error.message.decode('utf-8'))
        else:
            return OperationResult()


class GisdoUnitDictPack(UnitDictPack):

    url = r'/gisdo'
    _is_primary_for_model = False

    def get_nodes(self, parent_id, filter, branch_id=None):

        nodes = super(GisdoUnitDictPack, self).get_nodes(parent_id, filter, branch_id)
        try:
            result = [node for node in nodes if not node.gisdo.not_on_federal_report]
        except GisdoUnit.DoesNotExist:
            raise ApplicationLogicException(u'Учреждение c ID=%d не найдено в модуле "Фед. отчетности"' % node.pk)
        return result